
import matplotlib as mpl
import matplotlib.pylab as plt
import uuid
from . import logs

import trenchrun

import numpy as np
import numpy.ma as ma
from osgeo import gdal
gdal.UseExceptions()


def readBand(filename,
             bandNumber,
             npDataType = np.float32,
             normalize = False):

    ds = gdal.Open(filename)
    band = ds.GetRasterBand(bandNumber)
    projection = ds.GetProjection()
    array = band.ReadAsArray().astype(npDataType)
    if normalize:
        array = (array - array.min()) / (array.max() - array.min())

    # mask off any nodata values
    nodata = band.GetNoDataValue()
    logs.logger.info(f'nodata value: {nodata} for resource {filename}')
    nans = np.count_nonzero(array == nodata)
    array[array==nodata] = np.nan
    array = np.ma.array (array, mask=np.isnan(array))
    return array

def getInfo(filename):
    output = {}
    ds = gdal.Open(filename)
    projection = ds.GetProjection()
    output['projection'] = ds.GetProjection()
    output['metadata'] = ds.GetMetadata()
    output['transform'] = ds.GetGeoTransform()
    return output

class Blend(object):
    def __init__(self, data):
        self.data = data

    def do(self):
        intensityFilename = str(self.data.args.intensityPath)
        daylightFilename = str(self.data.args.aoPath)
        intensity = readBand(intensityFilename, 1, np.float32, True)
        daylight = readBand(daylightFilename, 1)

        logs.logger.info(f'Intensity shape {intensity.shape} ')
        logs.logger.info(f'Daylight shape {intensity.shape} ')

        intensity_mask = ma.getmask(intensity)
        daylight_mask = ma.getmask(daylight)
        if self.data.args.blue:
            cmap = mpl.cm.Blues_r
        else:
            cmap = mpl.cm.Greys_r

        intensity_RGBA = cmap(intensity)
        intensity_RGBA[...,3] = np.full(intensity.shape, self.data.args.alpha)

        cmap = mpl.cm.Greys_r
        daylight_RGB = cmap(daylight)

        RGBA = intensity_RGBA * 0.5 + daylight_RGB * 0.5

        numBands = RGBA.shape[2]


        nodata = 255
        big = (RGBA*255).astype(np.uint8)
        for i in range(numBands):
            ma.putmask(big[...,i], intensity_mask, nodata)
            ma.putmask(big[...,i], daylight_mask, nodata)

        tifpath = f"/vsimem/{str(uuid.uuid4())}.tif"
        gtif = gdal.GetDriverByName("GTiff")
        rast = gtif.Create(tifpath, intensity.shape[1], intensity.shape[0], numBands, gdal.GDT_Byte)

        info = getInfo(intensityFilename)

        rast.SetProjection(info['projection'])
        rast.SetGeoTransform(info['transform'])

        for b in range(numBands):
            band =rast.GetRasterBand(b+1)
            band.WriteArray(big[...,b])
            band.SetNoDataValue(nodata)

        png = gdal.GetDriverByName("PNG")

        description='daylight exposure mixed with lidar intensity'
        title = 'Absorptive Daylight Exposure'

        if self.data.args.full_output:
            png.CreateCopy( f"{self.data.args.output}-trenchrun.png", rast, 0,
                [ f'TITLE={title}', f'COMMENT={description}' ] )

        ds = gtif.CreateCopy( f"{self.data.args.output}-trenchrun.tif", rast, 0,
            [ 'COMPRESS=Deflate', 'TILED=YES','PREDICTOR=2' ] )

        if 'AREA_OR_POINT' in info['metadata']:
            ds.SetMetadataItem('AREA_OR_POINT', info['metadata']['AREA_OR_POINT'])

        ds.SetMetadataItem('TIFFTAG_SOFTWARE',f'Trenchrun {trenchrun.__version__}')
        ds.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION',f'{description}')
        ds.SetMetadataItem('TIFFTAG_DOCUMENTNAME',f'{title}')

        if self.data.args.full_output:
            ao = gdal.Open(str(self.data.args.aoPath))
            ds = gtif.CreateCopy( f"{self.data.args.output}-occlusion.tif", ao, 0,
                [ 'COMPRESS=Deflate', 'TILED=YES','PREDICTOR=2' ] )

            ao = gdal.Open(str(self.data.args.dsmPath))
            ds = gtif.CreateCopy( f"{self.data.args.output}-dsm.tif", ao, 0,
                [ 'COMPRESS=LZW', 'TILED=YES','PREDICTOR=3' ] )

            ao = gdal.Open(str(self.data.args.intensityPath))
            ds = gtif.CreateCopy( f"{self.data.args.output}-intensity.tif", ao, 0,
                [ 'COMPRESS=LZW', 'TILED=YES','PREDICTOR=2' ] )



from . import logs

import json
import tempfile
import subprocess
import shlex
import pathlib

import pdal


def run(command):
    args = shlex.split(command)
    p = subprocess.Popen(args,
                            stdin = subprocess.PIPE,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE,)
    ret = p.communicate()

    if p.returncode != 0:
        error = ret[1].decode('utf-8','replace')
        raise RuntimeError(error)

    response = ret[0].decode('utf-8','replace')
    return response

class Data(object):
    def __init__(self, args):

        self.args = args

        if '.json' in self.args.input.suffixes:
            self.inputType = 'pipeline'
        else:
            self.inputType = 'readable'

        self.args.intensityPath = pathlib.Path(tempfile.NamedTemporaryFile(suffix='.tif', delete=False).name)
        self.args.dsmPath = pathlib.Path(tempfile.NamedTemporaryFile(suffix='.tif', delete=False).name)
        self.args.aoPath = pathlib.Path(tempfile.NamedTemporaryFile(suffix='.tif', delete=False).name)
        self.checkValidData()

    def readPipeline(self):
        if self.inputType != 'pipeline':
            raise RuntimeError("Data type is not pipeline!")
        j = self.args.input.read_bytes().decode('utf-8')
        stages = pdal.pipeline._parse_stages(j)
        p = pdal.Pipeline(stages)

        # strip off any writers we're making our own
        stages = []
        for stage in p.stages:
            if stage.type.split('.')[0] != 'writers':
                stages.append(stage)

        p = pdal.Pipeline(stages)
        return p


    def readFile(self):
        reader = pdal.Reader(str(self.args.input))
        pipeline = reader.pipeline()
        return pipeline

    def __del__(self):
        self.args.intensityPath.unlink()
        self.args.dsmPath.unlink()
        self.args.aoPath.unlink()

    def checkValidData(self):

        if self.inputType == 'pipeline':
            reader = self.readPipeline()
        else:
            reader = self.readFile()

        qi = reader.quickinfo
        for key in qi:
            dimensions = [i.strip() for i in qi[key]['dimensions'].split(',')]
            if 'Intensity' not in dimensions:
                raise RuntimeError("Intensity information not available, this tool cannot run")

    def getReader(self):
        reader = None
        if self.args.reader_args:
            with open(self.args.reader_args,'r') as f:
                j = json.loads(f.read())
            reader = pdal.Reader(filename=str(self.args.input), *j)
        else:
            reader = pdal.Reader(filename=str(self.args.input))
        return reader

    def getWriters(self):
        intensity = pdal.Writer.gdal(
            filename=str(self.args.intensityPath),
            data_type='uint16_t',
            dimension='Intensity',
            output_type = 'idw',
            resolution=self.args.resolution,
        )
        dsm = pdal.Writer.gdal(
            filename=str(self.args.dsmPath),
            data_type='float',
            dimension='Z',
            output_type = 'idw',
            resolution=self.args.resolution,
        )
        return intensity | dsm

    def getPipeline(self):
        if self.inputType == 'pipeline':
            reader = self.readPipeline()
        else:
            reader = self.readFile()

        writers = self.getWriters()
        stage = reader | writers

        return stage


    def execute(self):
        pipeline = self.getPipeline()
        count = 0
        if pipeline.streamable:
            count = pipeline.execute_streaming(chunk_size=self.args.chunk_size)
        else:
            count = pipeline.execute()
        logs.logger.info(f'Wrote intensity and dsm for {count} points')

    def getImageCenter(self):
        # Run our pipeline

        command = f'gdalinfo -json {self.args.dsmPath}'
        response = run(command)
        j = json.loads(response)
        corner = j['wgs84Extent']['coordinates'][0][0]
        logs.logger.info(f'Fetched image center {corner}')
        return corner
    def ambient_occlusion(self):
        lng, lat = self.getImageCenter()

        command = f"""whitebox_tools -r=TimeInDaylight  \
        -i {self.args.dsmPath} -o {self.args.aoPath}  --az_fraction=15.0 \
        --max_dist=100.0 --lat={lat:.5f} --long={lng:.5f} """
        logs.logger.info(f"Processing ambient occlusion '{command}'")
        response = run(command)
        logs.logger.info(f"Processed ambient occlusion ")





from . import data
from . import blend
from .logs import logger 


def doIt(args):
    logger.info(f"opening {args.input} for processing")
    d = data.Data(args)
    d.execute()
    d.ambient_occlusion()

    b = blend.Blend(d)
    b.do()